# Copyright 2015 Infoblox Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from infoblox_client import exceptions as ib_ex
from infoblox_client import objects as obj
from infoblox_client import utils as ib_utils

LOG = logging.getLogger(__name__)


class InfobloxObjectManager(object):

    def __init__(self, connector):
        self.connector = connector

    def create_network_view(self, network_view, extattrs):
        return obj.NetworkView.create(self.connector,
                                      name=network_view,
                                      extattrs=extattrs)

    def delete_network_view(self, network_view):
        # never delete default network view
        if network_view == 'default':
            return
        nview = obj.NetworkView.search(self.connector,
                                       name=network_view)
        nview.delete()

    def create_dns_view(self, network_view, dns_view):
        return obj.DNSView.create(self.connector,
                                  name=dns_view,
                                  network_view=network_view)

    def delete_dns_view(self, dns_view):
        dns_view = obj.DNSView.search(self.connector,
                                      name=dns_view)
        dns_view.delete()

    def create_network(self, net_view_name, cidr, nameservers=None,
                       members=None, gateway_ip=None, dhcp_trel_ip=None,
                       network_extattrs=None):
        """Create NIOS Network."""

        options = []
        if nameservers:
            options.append(obj.DhcpOption(name='domain-name-servers',
                                          value=",".join(nameservers)))
        if gateway_ip:
            options.append(obj.DhcpOption(name='routers',
                                          value=gateway_ip))
        if dhcp_trel_ip:
            options.append(obj.DhcpOption(name='dhcp-server-identifier',
                                          num=54,
                                          value=dhcp_trel_ip))
        return obj.Network.create(self.connector,
                                  network_view=net_view_name,
                                  cidr=cidr,
                                  members=members,
                                  options=options,
                                  extattrs=network_extattrs,
                                  check_if_exists=False)

    def get_network(self, network_view, cidr):
        network = obj.Network.search(self.connector,
                                     network_view=network_view,
                                     cidr=cidr)
        if not network:
            raise ib_ex.InfobloxNetworkNotAvailable(
                network_view=network_view, cidr=cidr)
        return network

    def create_ip_range(self, network_view, start_ip, end_ip, network,
                        disable, range_extattrs):
        return obj.IPRange.create(self.connector,
                                  network_view=network_view,
                                  start_addr=start_ip,
                                  end_addr=end_ip,
                                  cidr=network,
                                  disable=disable,
                                  extattrs=range_extattrs)

    def delete_ip_range(self, network_view, start_ip, end_ip):
        range = obj.IPRange.search(self.connector,
                                   network_view=network_view,
                                   start_addr=start_ip,
                                   end_addr=end_ip)
        range.delete()

    def has_networks(self, network_view_name):
        try:
            networks = obj.Network.search_all(self.connector,
                                              network_view=network_view_name)
            return bool(networks)
        except ib_ex.InfobloxSearchError:
            return False

    def network_exists(self, network_view, cidr):
        try:
            network = obj.Network.search(self.connector,
                                         network_view=network_view,
                                         cidr=cidr)
        except ib_ex.InfobloxSearchError:
            network = None
        return network is not None

    def delete_network(self, network_view, cidr):
        network = obj.Network.search(self.connector,
                                     network_view=network_view,
                                     cidr=cidr)
        network.delete()

    def create_network_from_template(self, network_view, cidr, template,
                                     extattrs):
        return obj.Network.create(self.connector,
                                  network_view=network_view,
                                  cidr=cidr,
                                  template=template,
                                  extattrs=extattrs,
                                  check_if_exists=False)

    def update_network_options(self, ib_network, extattrs=None):
        if extattrs:
            ib_network.extattrs = extattrs
        return ib_network.update()

    def get_host_record(self, dns_view, ip):
        return obj.HostRecord.search(self.connector,
                                     view=dns_view,
                                     ip=ip)

    def find_hostname(self, dns_view, hostname, ip):
        return obj.HostRecord.search(self.connector,
                                     name=hostname,
                                     view=dns_view,
                                     ip=ip)

    def create_host_record_for_given_ip(self, dns_view, zone_auth,
                                        hostname, mac, ip, extattrs,
                                        use_dhcp):
        name = '.'.join([hostname, zone_auth])
        ip_obj = obj.IP.create(ip=ip, mac=mac, configure_for_dhcp=use_dhcp)
        return obj.HostRecord.create(self.connector,
                                     view=dns_view,
                                     name=name,
                                     ip=ip_obj,
                                     extattrs=extattrs,
                                     check_if_exists=False)

    def create_host_record_from_range(self, dns_view, network_view_name,
                                      zone_auth, hostname, mac, first_ip,
                                      last_ip, extattrs, use_dhcp):
        name = '.'.join([hostname, zone_auth])
        ip_alloc = obj.IPAllocation.next_available_ip_from_range(
            network_view_name, first_ip, last_ip)
        ip_obj = obj.IP.create(ip=ip_alloc, mac=mac,
                               configure_for_dhcp=use_dhcp)
        return obj.HostRecord.create(self.connector,
                                     view=dns_view,
                                     name=name,
                                     ip=ip_obj,
                                     extattrs=extattrs,
                                     check_if_exists=False)

    def delete_host_record(self, dns_view, ip_address):
        host_record = obj.HostRecord.search(self.connector,
                                            view=dns_view, ip=ip_address)
        host_record.delete()

    def create_fixed_address_for_given_ip(self, network_view, mac, ip,
                                          extattrs):
        return obj.FixedAddress.create(self.connector,
                                       network_view=network_view,
                                       mac=mac,
                                       ip=ip,
                                       extattrs=extattrs,
                                       check_if_exists=False)

    def create_fixed_address_from_range(self, network_view, mac, first_ip,
                                        last_ip, extattrs):
        ip = obj.IPAllocation.next_available_ip_from_range(
            network_view, first_ip, last_ip)
        return obj.FixedAddress.create(self.connector,
                                       ip=ip,
                                       mac=mac,
                                       network_view=network_view,
                                       extattrs=extattrs,
                                       check_if_exists=False)

    def create_fixed_address_from_cidr(self, netview, mac, cidr, extattrs):
        ip = obj.IPAllocation.next_available_ip_from_cidr(netview, cidr)
        return obj.FixedAddress.create(self.connector,
                                       network_view=netview,
                                       ip=ip,
                                       mac=mac,
                                       extattrs=extattrs,
                                       check_if_exists=False)

    def delete_fixed_address(self, network_view, ip_address):
        fixed_address = obj.FixedAddress.search(self.connector,
                                                network_view=network_view,
                                                ip=ip_address)
        fixed_address.delete()

    def add_ip_to_record(self, host_record, ip, mac, use_dhcp=True):
        ip_obj = obj.IP.create(ip=ip, mac=mac, configure_for_dhcp=use_dhcp)
        host_record.ip.append(ip_obj)
        return host_record.update()

    def add_ip_to_host_record_from_range(self, host_record, network_view,
                                         mac, first_ip, last_ip,
                                         use_dhcp=True):
        ip_alloc = obj.IPAllocation.next_available_ip_from_range(
            network_view, first_ip, last_ip)
        ip_obj = obj.IP.create(ip=ip_alloc, mac=mac,
                               configure_for_dhcp=use_dhcp)
        host_record.ip.append(ip_obj)
        return host_record.update()

    def delete_ip_from_host_record(self, host_record, ip):
        host_record.ip.remove(ip)
        return host_record.update()

    def has_dns_zones(self, dns_view):
        try:
            zones = obj.DNSZone.search_all(self.connector, view=dns_view)
            return bool(zones)
        except ib_ex.InfobloxSearchError:
            return False

    def create_dns_zone(self, dns_view, dns_zone,
                        grid_primary=None, grid_secondaries=None,
                        zone_format=None, ns_group=None, prefix=None,
                        extattrs=None):
        try:
            return obj.DNSZone.create(self.connector,
                                      fqdn=dns_zone,
                                      view=dns_view,
                                      extattrs=extattrs,
                                      zone_format=zone_format,
                                      ns_group=ns_group,
                                      prefix=prefix,
                                      grid_primary=grid_primary,
                                      grid_secondaries=grid_secondaries)
        except ib_ex.InfobloxCannotCreateObject:
            LOG.warning('Unable to create DNS zone %(dns_zone_fqdn)s '
                        'for %(dns_view)s',
                        {'dns_zone_fqdn': dns_zone, 'dns_view': dns_view})

    def delete_dns_zone(self, dns_view, dns_zone_fqdn):
        dns_zone = obj.DNSZone.search(self.connector,
                                      fqdn=dns_zone_fqdn,
                                      view=dns_view)
        dns_zone.delete()

    def update_host_record_eas(self, dns_view, ip, extattrs):
        host_record = obj.HostRecord.search(self.connector,
                                            view=dns_view,
                                            ip=ip)
        host_record.extattrs = extattrs
        host_record.update()

    def update_fixed_address_eas(self, network_view, ip, extattrs):
        fixed_address = obj.FixedAddress.search(self.connector,
                                                network_view=network_view,
                                                ip=ip)
        fixed_address.extattrs = extattrs
        fixed_address.update()

    def update_dns_record_eas(self, dns_view, ip, extattrs):
        a_record = obj.ARecordBase.search(self.connector,
                                          ip=ip,
                                          view=dns_view)
        if a_record:
            a_record.extattrs = extattrs
            a_record.update()

        ptr_record = obj.PtrRecord.search(self.connector,
                                          ip=ip,
                                          view=dns_view)
        if ptr_record:
            ptr_record.extattrs = extattrs
            ptr_record.update()

    def bind_name_with_host_record(self, dns_view, ip, name, extattrs):
        host_record = obj.HostRecord.search(self.connector,
                                            view=dns_view,
                                            ip=ip)
        if host_record:
            host_record.name = name
            host_record.extattrs = extattrs
            host_record.update()

    def bind_name_with_record_a(self, dns_view, ip, name, bind_list,
                                extattrs):
        is_ipv4 = ib_utils.determine_ip_version(ip) == 4
        if ((is_ipv4 and 'record:a' in bind_list) or
                (not is_ipv4 and 'record:aaaa' in bind_list)):
            obj.ARecordBase.create(self.connector,
                                   view=dns_view,
                                   ip=ip,
                                   name=name,
                                   extattrs=extattrs)

        if 'record:ptr' in bind_list:
            obj.PtrRecord.create(self.connector,
                                 view=dns_view,
                                 ip=ip,
                                 ptrdname=name,
                                 extattrs=extattrs)

    def unbind_name_from_record_a(self, dns_view, ip, name, unbind_list):
        is_ipv4 = ib_utils.determine_ip_version(ip) == 4
        if ((is_ipv4 and 'record:a' in unbind_list) or
                (not is_ipv4 and 'record:aaaa' in unbind_list)):
            a_record = obj.ARecordBase.search(self.connector,
                                              view=dns_view,
                                              ip=ip,
                                              name=name)
            a_record.delete()

        if 'record:ptr' in unbind_list:
            ptr_record = obj.PtrRecord.search(self.connector,
                                              view=dns_view,
                                              ip=ip,
                                              ptrdname=name)
            ptr_record.delete()

    def get_member(self, member):
        member.fetch()
        return member

    def get_all_ea_definitions(self):
        try:
            ea_defs = obj.EADefinition.search_all(self.connector)
            return ea_defs
        except ib_ex.InfobloxSearchError:
            return None

    def create_ea_definition(self, ea_def):
        try:
            return obj.EADefinition.create(self.connector,
                                           check_if_exists=False,
                                           **ea_def)
        except ib_ex.InfobloxCannotCreateObject:
            LOG.error('Unable to create Extensible Attribute Definition '
                      '%s' % ea_def)

    def create_required_ea_definitions(self, required_ea_defs):
        existing_ea_defs = self.get_all_ea_definitions()
        missing_ea_defs = filter(lambda x: not next(
            (y for y in existing_ea_defs if x['name'] == y.name), None),
            required_ea_defs)

        for ea_def in missing_ea_defs:
            self.create_ea_definition(ea_def)

    def restart_all_services(self, member):
        if not member._ref:
            member.fetch()
        self.connector.call_func('restartservices', member._ref,
                                 {'restart_option': 'RESTART_IF_NEEDED',
                                  'service_option': 'ALL'})

    def get_object_refs_associated_with_a_record(self, a_record_ref):
        # record should in the format: {object_type, search_field}
        associated_with_a_record = [
            {'type': 'record:cname', 'search': 'canonical'},
            {'type': 'record:txt', 'search': 'name'}
        ]

        ib_obj_refs = []
        a_record = self.connector.get_object(a_record_ref)

        for rec_inf in associated_with_a_record:
            obj_type = rec_inf['type']
            payload = {'view': a_record['view'],
                       rec_inf['search']: a_record['name']}
            ib_objs = self.connector.get_object(obj_type, payload)
            if ib_objs:
                for ib_obj in ib_objs:
                    ib_obj_refs.append(ib_obj['_ref'])
        return ib_obj_refs

    def get_all_associated_objects(self, network_view, ip):
        ip_objects = obj.IPAddress.search(self.connector,
                                          network_view=network_view,
                                          ip_address=ip)
        if ip_objects:
            return ip_objects.objects
        return []

    @staticmethod
    def _get_object_type_from_ref(ref):
        return ref.split('/', 1)[0]

    def delete_all_associated_objects(self, network_view, ip, delete_list):
        del_ib_objs = []
        ib_obj_refs = self.get_all_associated_objects(network_view, ip)

        for ib_obj_ref in ib_obj_refs:
            del_ib_objs.append(ib_obj_ref)
            obj_type = self._get_object_type_from_ref(ib_obj_ref)
            if obj_type in ['record:a', 'record:aaaa']:
                del_ib_objs.extend(
                    self.get_object_refs_associated_with_a_record(ib_obj_ref))

        for ib_obj_ref in del_ib_objs:
            obj_type = self._get_object_type_from_ref(ib_obj_ref)
            if obj_type in delete_list:
                self.connector.delete_object(ib_obj_ref)

    def delete_object_by_ref(self, ref):
        try:
            self.connector.delete_object(ref)
        except ib_ex.InfobloxCannotDeleteObject:
            pass



# Copyright 2015 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
import decimal
import netaddr
import random
import re
import six

from oslo_log import log as logging
from oslo_serialization import jsonutils


LOG = logging.getLogger(__name__)
EA_CLOUD_MGMT_PLATFORM_TYPE = 'CMP Type'


def json_to_obj(obj_type, json_data):
    """Converts json data to an object with a given object type

    :param obj_type: converted object's type that is determined dynamically
    :param json_data: json string or json object
    :return: object
    """
    def dic2obj(x):
        if isinstance(x, dict):
            return type(obj_type, (),
                        {k: dic2obj(v) for k, v in six.iteritems(x)})
        else:
            return x

    if isinstance(json_data, six.string_types):
        json_data = jsonutils.loads(json_data)
    return dic2obj(json_data)


def get_values_from_records(key, records):
    key_vals = []
    if records is None:
        return key_vals

    for record in records:
        key_val = record.get(key, None)
        if key_val:
            key_vals.append(key_val)
    return key_vals


def db_records_to_obj(obj_type, records):
    record_json = db_records_to_json(records)
    if not isinstance(record_json, list):
        return json_to_obj(obj_type, record_json)

    result_set = []
    for record in records:
        result_set.append(json_to_obj(obj_type, record))
    return result_set


def db_records_to_json(records):
    """Converts db records to json.

    alchemy_encoder is needed for date and numeric(x,y) fields since
    they will turn into datetime.date and decimal.Decimal types.
    """
    def alchemy_encoder(obj):
        if isinstance(obj, datetime.date):
            return obj.isoformat()
        elif isinstance(obj, decimal.Decimal):
            return float(obj)

    rows = []
    for record in records:
        if isinstance(record, tuple):
            merge = dict()
            for table in record:
                merge.update(dict(table))
            rows.append(merge)
        else:
            rows.append(dict(record))

    # return all rows as a JSON array of objects
    json_str = jsonutils.dumps(rows, alchemy_encoder)
    return jsonutils.loads(json_str)


def construct_ea(attributes):
    ea = {}
    for name, value in six.iteritems(attributes):
        str_val = get_string_or_none(value)
        if str_val:
            ea[name] = {'value': str_val}

    ea[EA_CLOUD_MGMT_PLATFORM_TYPE] = {'value': 'OpenStack'}
    return ea


def get_string_or_none(value):
    ret_val = None
    if isinstance(value, six.string_types):
        ret_val = value
    else:
        if value is not None:
            ret_val = str(value)
    return ret_val


def scalar_from_ea(name, extattrs):
    valid = (name and isinstance(name, six.string_types) and
             extattrs and isinstance(extattrs, dict))
    if not valid:
        ValueError("Invalid argument was passed")

    value = None
    if extattrs:
        root = extattrs.get("extattrs")
        name_attr = root.get(name)
        if name_attr:
            value = name_attr.get('value')
    return value


def get_ip_version(ip_address):
    valid = ip_address and isinstance(ip_address, six.string_types)
    if not valid:
        ValueError("Invalid argument was passed")

    if type(ip_address) is dict:
        ip = ip_address['ip_address']
    else:
        ip = ip_address

    try:
        ip = netaddr.IPAddress(ip)
    except ValueError:
        ip = netaddr.IPNetwork(ip)
    return ip.version


def is_valid_ip(ip):
    try:
        netaddr.IPAddress(ip)
    except netaddr.core.AddrFormatError:
        return False
    return True


def generate_duid(mac):
    """DUID is consisted of 10 hex numbers.

    0x00 + 3 random hex + mac with 6 hex
    """
    valid = mac and isinstance(mac, six.string_types)
    if not valid:
        ValueError("Invalid argument was passed")
    duid = [0x00,
            random.randint(0x00, 0x7f),
            random.randint(0x00, 0xff),
            random.randint(0x00, 0xff)]
    return ':'.join(map(lambda x: "%02x" % x, duid)) + ':' + mac


def get_prefix_for_dns_zone(subnet_name, cidr):
    valid = cidr and isinstance(cidr, six.string_types)
    if not valid:
        ValueError("Invalid argument was passed")

    subnet_name = subnet_name if subnet_name else ''
    try:
        ip_version = get_ip_version(cidr)
    except netaddr.core.AddrFormatError:
        return ValueError("Invalid cidr")

    # add prefix only for classless networks (ipv4) mask greater than
    # 24 needs prefix; use meaningful prefix if used
    prefix = None
    if ip_version == 4:
        m = re.search(r'/\d+', cidr)
        mask = m.group().replace("/", "")
        if int(mask) > 24:
            if len(subnet_name) > 0:
                prefix = subnet_name
            else:
                prefix = '-'.join(
                    filter(None,
                           re.split(r'[.:/]', cidr))
                )
    return prefix


def get_physical_network_meta(network):
    valid = network and isinstance(network, dict)
    if not valid:
        ValueError("Invalid argument was passed")

    network = network if network else {}
    provider_network_type = network.get('provider:network_type')
    provider_physical_network = network.get('provider:physical_network')
    provider_segmentation_id = network.get('provider:segmentation_id')
    network_meta = {'network_type': provider_network_type,
                    'physical_network': provider_physical_network,
                    'segmentation_id': provider_segmentation_id}
    return network_meta


def get_list_from_string(data_string, delimiter_list):
    valid = (data_string and
             isinstance(data_string, six.string_types) and
             delimiter_list and
             isinstance(delimiter_list, list))
    if not valid:
        ValueError("Invalid argument was passed")

    list_data = remove_any_space(data_string)
    if isinstance(delimiter_list, six.string_types):
        if len(delimiter_list) == 0:
            return data_string
        return list_data.split(delimiter_list)

    if isinstance(delimiter_list, list):
        delimiter_count = len(delimiter_list)
        if delimiter_count == 0:
            return data_string
        if delimiter_count == 1:
            return list_data.split(delimiter_list[0])
        if delimiter_count > 2:
            return ValueError("Delimiter list can contain up to 2 delimiters.")

        result_list = []
        for delimiter in delimiter_list:
            if isinstance(list_data, six.string_types):
                list_data = list_data.split(delimiter)
            else:
                for ld in list_data:
                    result_list.append(ld.split(delimiter))
        # clean up empty string element ['']
        result_list[0] = [m for m in result_list[0] if m]
        result_list[1] = [m for m in result_list[1] if m]
        return result_list

    return ValueError("Unsupported delimiter list type")


def exists_in_sequence(sub_sequence_to_find, full_list_in_sequence):
    valid = (sub_sequence_to_find and
             isinstance(sub_sequence_to_find, list) and
             full_list_in_sequence and
             isinstance(full_list_in_sequence, list))
    if not valid:
        ValueError("Invalid argument was passed")

    return any(full_list_in_sequence[pos:pos + len(sub_sequence_to_find)] ==
               sub_sequence_to_find for pos in
               range(0,
                     len(full_list_in_sequence) - len(sub_sequence_to_find)
                     + 1))


def exists_in_list(list_to_find, full_list):
    valid = (list_to_find and isinstance(list_to_find, list) and
             full_list and isinstance(full_list, list))
    if not valid:
        ValueError("Invalid argument was passed")

    found_list = [m for m in list_to_find if m in full_list]
    return len(found_list) == len(list_to_find)


def find_one_in_list(search_key, search_value, search_list):
    valid = (search_key and isinstance(search_key, six.string_types) and
             search_value and isinstance(search_value, six.string_types) and
             search_list and isinstance(search_list, list))
    if not valid:
        ValueError("Invalid argument was passed")

    found_list = [m for m in search_list
                  if m.get(search_key) == search_value]
    return found_list[0] if found_list else None


def find_in_list(search_key, search_values, search_list):
    valid = (search_key and isinstance(search_key, six.string_types) and
             search_values and isinstance(search_values, list) and
             search_list and isinstance(search_list, list))
    if not valid:
        ValueError("Invalid argument was passed")

    found_list = [m for m in search_list
                  if m.get(search_key) in search_values]
    return found_list


def merge_list(*list_args):
    merge_lsit = []
    for lst in list_args:
        merge_lsit += lst
    return list(set(merge_lsit))


def remove_any_space(text):
    return re.sub(r'\s+', '', text)


def determine_ip_version(ip_in):
    ip_ver = 4
    if isinstance(ip_in, (list, tuple)):
        ip_in = ip_in[0]
    if ip_in:
        if isinstance(ip_in, int):
            if ip_in == 6:
                ip_ver = 6
            else:
                ip_ver = 4
        elif hasattr(ip_in, 'ip_version'):
            return ip_in.ip_version
        else:
            if type(ip_in) is dict:
                addr = ip_in['ip_address']
            else:
                addr = ip_in
            try:
                ip = netaddr.IPAddress(addr)
            except ValueError:
                ip = netaddr.IPNetwork(addr)

            ip_ver = ip.version
    return ip_ver
